/**
  * Nazwa: boost_thread.cpp
  * Autor: Micha Ulaski	(F1-TIZTK)
  * Data: 11-03-07 13:37
  * Opis: Demonstracja i zastosowanie biblioteki boost::thread
  * Zakres:
            + tworzenie wtku na bazie obiektw
            + tworzenie wtkw na bazie funkcji
            + mechanizm usypiania wtkw
            + uywanie thread specific ptr
  */

#define VERSION "0.2.1"

#include <iostream>
#include <vector>
#include <boost/thread.hpp>

/**
 * Klasa Obiekt:
 * Jest to prosty licznik
 * Metoda run() jest wykorzystywana do utworzenia wtku,
 * jednak jedynym moliwym sposobem uczynienia tego
 * jest zastosowanie wzorca komendy
 */
class Object
{
    private:

        std::string t_name;     ///< nazwa "wtku"
        bool t_run;             ///< warto logiczna uywana do tworzenia wtku
        int t_counter;          ///< licznik

    public:

        Object(std::string t_name)
        {
            this->t_name = t_name;
            this->t_run = false;
            this->t_counter = 0;
        }

        ~Object()
        {
        }

        std::string get_name() const
        {
            return this->t_name;
        }
        
        bool get_run() const
        {
            return this->t_run;
        }

        void set_run(bool t_run)
        {
            this->t_run = t_run;
        }

        /**
         * Metoda run()
         * To na jej podstawie jest tworzony wtek przy uyciu funktora
         */
        int run()
        {
            boost::xtime xt;    // jest to obiekt sucy operacji zwizanych z czasem (o wysokiej precyzji) [jest to jednak rozwizanie tymczasowe i w kolejnych wersjach biblioteki zostanie zastpione]

            while(this->t_run)
            {
                this->t_counter++;
                std::cout << *this << std::endl;
                boost::xtime_get(&xt, boost::TIME_UTC);   // w ten sposb "wrzucamy" do powyszego obiektu biecy czas (jako okres od pocztku epoki), zgodny z podanym zegarem
                xt.nsec += 500000000;                     // moemy dowolnie zwikszy warto (w tym przypadku o p sekundy)
                boost::thread::sleep(xt);                 // a nastpnie "uspi" wtek do momentu gdy nadejdzie odpowiedni czas :)
            }
            return t_counter;
        };

        friend std::ostream& operator<<(std::ostream& out, const Object& o);
};

std::ostream& operator<<(std::ostream& out, const Object& o)
{
    out << "NAZWA: " << o.t_name << " LICZNIK: " << o.t_counter;
    return out;
};

/**
 * thread_specific_ptr
 * jest to specjalny mechanizm umoliwajcy
 * przechowywanie odrbnych danych tego samego typu
 * w kadym z wtkw
 * to znaczy, kady wtek przy odwoywaniu si do utworzonego
 * wskanika na int (w tym take zapisywaniu nowej wartoci)
 * bdzie tak naprawd dziaa na innych danych
 * dane zapisane przez wtek s czyszczone gdy si on zakoczy
 */
boost::thread_specific_ptr<int> count;

/**
 * Klasa Funktor:
 * Jest to wzorzec komendy, suy on do tworzenia wtkw na bazie metod
 * (a w zaszadzie metody run()) danego obiektu
 */
class Functor
{
    private:

        Object* object;

    public:

        Functor(Object* object)
        {
            this->object = object;
        }

        ~Functor()
        {
        }

        void operator()()
        {
            count.reset(new int(0));   // sposb inicjalizacji thread_specific_ptr dla wtku
            if(object)
            {
                *count += object->run();   // gdyby nie bya to odrbna pami dla kadego wtku, to przy zakoczeniu ostatniego wtku mieli bymy warto inn od faktycznej wartoci licznika
                std::cout << std::endl << "\t\tNAZWA: " << object->get_name() << " DOLICZY DO: " << *count << std::endl << std::endl;
            }
            return;
        };
};

/**
 * Wektor wskanikw na obiekty wykorzystywane w funkcji
 * threads_monitor
 */
std::vector<Object*> objects;

/**
 * Wtki mona take tworzy na bazie zwykych funkcji.
 * Tak si wanie dzieje z funkcja threads_monitor()
 * Sprawdza ona stan wszystkich wtkw (objects)
 * i w przypadku, gdy wszystkie wtki si ju zakonczyy
 * koczy dziaanie.
 */
void threads_monitor()
{
    boost::xtime xt;    // mechanizm uzypiania wtku, analogiczny jak w metodzie run() klasy obiekt
    bool state;
    while(true)
    {
        state = false;
        std::cout << "\t\t\t\tMONITOR:" << std::endl;
        for(int i = 0; i < objects.size(); i++)
        {
            std::cout << "\t\t\t\t" << *(objects.at(i)) << "  --STATE: ";
            if(objects.at(i)->get_run())
            {
                std::cout << "active" << std::endl;
                state = true;
            }
            else
            {
                std::cout << "inactive" << std::endl;
            }
        }
        if(!state)
        {
            break;
        }
        else
        {
            boost::xtime_get(&xt, boost::TIME_UTC);
            xt.nsec += 1500000000;
            boost::thread::sleep(xt);
        }
    }
    return;
}

int main(int argc,char *argv[])
{
    Object object_1("wtek_1");
    objects.push_back(&object_1);
    Functor functor_1(&object_1);
    object_1.set_run(true);
    boost::thread thread_1(functor_1);  // tworzenie nowego wtku na bazie obiektu, przy uyciu funktora

    Object object_2("wtek_2");
    objects.push_back(&object_2);
    Functor functor_2(&object_2);
    object_2.set_run(true);
    boost::thread thread_2(functor_2);  // j.w.

    Object object_3("wtek_3");
    objects.push_back(&object_3);
    Functor functor_3(&object_3);
    object_3.set_run(true);
    boost::thread thread_3(functor_3);  // j.w.

    boost::thread monitor(&threads_monitor);  // tworzenie nowego wtku na bazie zwykej funkcji

    sleep(5);
    std::cout << std::endl << "\t\t!!!WYCZAMY 1 WTEK!!!" << std::endl << std::endl;
    object_1.set_run(false);
    sleep(3);
    std::cout << std::endl << "\t\t!!!WYCZAMY 3 WTEK!!!" << std::endl << std::endl;
    object_3.set_run(false);
    sleep(2);
    std::cout << std::endl << "\t\t!!!WYCZAMY 2 WTEK!!!" << std::endl << std::endl;
    object_2.set_run(false);
    monitor.join();     // wywoanie funkcji join powoduje, zatrzymanie biecego wtku do momentu zakoczenia si wtku dla ktrego wywoalimy join

    return 0;
}

//
